/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.xml;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.text.CollationKey;
import java.text.Collator;
import java.text.FieldPosition;
import java.text.Format;
import java.util.Map;
import java.util.Map.Entry;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Set;
import java.util.regex.Pattern;
import java.util.zip.DataFormatException;
import java.util.zip.Deflater;
import java.util.zip.Inflater;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import com.topcoder.document.index.CollectionIndex;
import com.topcoder.document.index.DocumentIndex;
import com.topcoder.document.index.persistence.IndexPersistence;
import com.topcoder.document.index.persistence.IndexPersistenceException;
import com.topcoder.document.index.persistence.impl.IndexPersistenceHelper;
import com.topcoder.document.index.persistence.impl.PersistenceConfigurationException;
import com.topcoder.document.index.wordsource.WordSourceId;
import com.topcoder.util.compression.Base64Codec;
import com.topcoder.util.compression.Base64Decoder;
import com.topcoder.util.compression.Base64Encoder;
import com.topcoder.util.config.ConfigManager;
import com.topcoder.util.config.ConfigManagerException;

/**
 * Implements <code>IndexPersistence</code> interface to provide the functionality to persist document indices and
 * collection indices as XML files. The source identity instance of the document index must be serializable. The
 * structure of the XML files are described as follows.<br>
 * One document index is persisted as two parts: several document index files containing the document ID, words and
 * their positions, and one element in the 'document_usage.xml' file containing the document usage count and collection
 * IDs which the document index belongs to. The document index files are named 'doc_index_{id}_{sequence_id}.xml'.
 * '{id}' means the string ID of the document index. '{sequence_id}' is a set of consecutive integers starting from 1.
 * The document ID information is persisted in every document index files. For one single word, its positions also only
 * exist in one file. However, different words of the same document index may spread across the files.<br>
 * One collection index is persisted in a single XML file. It only contains the collection ID and all document string
 * IDs in the collection index. The file name is formed as 'doc_collection_{collection_id}.xml'.<br>
 * Here are examples of the three kinds of XML files.<br>
 * document_usage.xml:
 *
 * <pre>
 *  &lt;?xml version=&quot;1.0&quot; encoding=&quot;UTF-8&quot;?&gt;
 *  &lt;documents&gt;
 *   &lt;document-usage document-id=&quot;document string ID&quot; usage-count=&quot;2&quot;&gt;
 *    &lt;collection-inclusion&gt;
 *     &lt;collection collection-id=&quot;collection ID 1&quot;/&gt;
 *     &lt;collection collection-id=&quot;collection ID 2&quot;/&gt;
 *     ...
 *    &lt;/collection-inclusion&gt;
 *   &lt;/document-usage&gt;
 *   ...
 *  &lt;/documents&gt;
 * </pre>
 *
 * doc_index_id_1.xml:
 *
 * <pre>
 *  &lt;?xml version=&quot;1.0&quot; encoding=&quot;UTF-8&quot;?&gt;
 *  &lt;doc xml:lang=&quot;en-US&quot;&gt;
 *   &lt;document-index document-id=&quot;document string ID&quot;&gt;
 *    &lt;word-source-data&gt;
 *     &lt;source-identity&gt;BASE64 encoded serialized document source instance&lt;/source-identity&gt;
 *     &lt;class-name&gt;com.topcoder.document.index.wordsource.textfile.TextFileWordSource&lt;/class-name&gt;
 *     &lt;locale language=&quot;en&quot; country=&quot;US&quot; variant=&quot;&quot;/&gt;
 *     &lt;delimiters&gt;
 *      &lt;entry delimiter=&quot;delimiter 1&quot;/&gt;
 *      &lt;entry delimiter=&quot;delimiter 2&quot;/&gt;
 *      ...
 *     &lt;/delimiters&gt;
 *    &lt;/word-source-data&gt;
 *    &lt;word-index&gt;
 *     &lt;word&gt;Word1&lt;/word&gt;
 *     &lt;pos&gt;1&lt;/pos&gt;
 *     &lt;pos&gt;12&lt;/pos&gt;
 *     ...
 *    &lt;/word-index&gt;
 *    ...
 *   &lt;/document-index&gt;
 *  &lt;/doc&gt;
 * </pre>
 *
 * doc_collection_id.xml:
 *
 * <pre>
 *  &lt;?xml version=&quot;1.0&quot; encoding=&quot;UTF-8&quot;?&gt;
 *  &lt;document-collection collection-id=&quot;collection ID&quot;&gt;
 *   &lt;documents&gt;
 *    &lt;document-id&gt;document string ID 1&lt;/document-id&gt;
 *    &lt;document-id&gt;document string ID 2&lt;/document-id&gt;
 *    ...
 *   &lt;/documents&gt;
 *  &lt;/document-collection&gt;
 * </pre>
 *
 * When persisting the delimiters and word data, XML illegal characters are escaped in the format of '&#92;uXXXX'.
 * 'XXXX' is the four-digit hex string of the unicode character.<br>
 * The document usage file "document_usage.xml" must exist in the persistence base directory.<br>
 * This class can be configured by configuration manager. Under the given namespace, three property
 * 'XmlPersistencePath', 'FileSizeLimit', and 'FileSizeLimit' must exist. Property 'XmlPersistencePath' defines an
 * existing directory used as the base directory of all XML files to be accessed. Property 'FileSizeLimit' defines a
 * positive integer, representing the maximum size of the document index files. It is in mega-bytes (1024*1024 bytes).
 * Property 'FileSizeThreshold' defines a non-negative integer, representing the safety margin of the file size. It is
 * in kilo-bytes (1024 bytes). When a document index file being written exceeds the maximum size subtracting the safety
 * margin, the document index file is ended, and a new document index with the next sequence ID is created.<br>
 * This class is immutable. However, it is not thread-safe, since it access some external files.
 *
 * @author AleaActaEst, TCSDEVELOPER
 * @version 1.0
 */
public class XmlIndexPersistence implements IndexPersistence {
    /**
     * Represents the property name where the path of the XML files is loaded.
     */
    private static final String XML_PATH_PROPERTY = "XmlPersistencePath";

    /**
     * Represents the property name where the maximum size of the XML file is loaded.
     */
    private static final String FILE_SIZE_LIMIT_PROPERTY = "FileSizeLimit";

    /**
     * Represents the property name where the file size safety margin of the XML file is loaded.
     */
    private static final String FILE_SIZE_THRESHOLD_PROPERTY = "FileSizeThreshold";

    /**
     * Represents the element name where a group of document indices are stored.
     */
    private static final String DOCUMENTS_ELEMENT = "documents";

    /**
     * Represents the escaping character used to escape illegal XML characters.
     */
    private static final char ESCAPE_CHARACTER = '\\';

    /**
     * Represents the name of the UTF-8 encoding.
     */
    private static final String ENCODING = "UTF-8";

    /**
     * Represents the file extension of the XML file.
     */
    private static final String XML_EXTENSION = ".xml";

    /**
     * Represents the regular expression used to match the sequence ID part of the XML file containing document indices.
     */
    private static final String DOC_INDEX_SEQUENCE_PATTERN = "_(\\d+)\\.xml";

    /**
     * Represents the XML escaping prefix of a character.
     */
    private static final String ESCAPE_CHARACTER_PREFIX = "&#";

    /**
     * Represents an empty string.
     */
    private static final String EMPTY = "";

    /**
     * Represents the number of bytes in 1 kilo-bytes.
     */
    private static final long KILO = 1024;

    /**
     * Represents the XML file header.
     */
    private static final String XML_HEADER = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>";

    /**
     * Represents the beginning of the XML file containing one document index.
     */
    private static final Format DOC_INDEX_XML_DOCUMENT_HEAD = new MessageFormat("<doc xml:lang=\"{0}\">"
        + "<document-index document-id=\"{1}\"><word-source-data><source-identity>{2}</source-identity>"
        + "<class-name>{3}</class-name><locale language=\"{4}\" country=\"{5}\" variant=\"{6}\"/><delimiters>");

    /**
     * Represents the end of the delimiters in the XML file.
     */
    private static final String DELIMITERS_END = "</delimiters></word-source-data>";

    /**
     * Represents the openning tag of the delimiter entry.
     */
    private static final String DELIMITER_ENTRY_OPEN = "<entry delimiter=\"";

    /**
     * Represents the closing tag of the delimiter entry.
     */
    private static final String DELIMITER_ENTRY_CLOSE = "\"/>";

    /**
     * Represents the end of the XML file containing document indices.
     */
    private static final String DOC_INDEX_XML_DOCUMENT_END = "</document-index></doc>";

    /**
     * Represents the openning tag of the word index element.
     */
    private static final String WORD_INDEX_OPEN = "<word-index>";

    /**
     * Represents the closing tag of the word index element.
     */
    private static final String WORD_INDEX_CLOSE = "</word-index>";

    /**
     * Represents the openning tag of the word element.
     */
    private static final String WORD_OPEN = "<word>";

    /**
     * Represents the closing tag of the word element.
     */
    private static final String WORD_CLOSE = "</word>";

    /**
     * Represents the openning tag of the position element.
     */
    private static final String POSITION_OPEN = "<pos>";

    /**
     * Represents the closing tag of the position element.
     */
    private static final String POSITION_CLOSE = "</pos>";

    /**
     * Represents the element name where the document index usage count is stored.
     */
    private static final String DOCUMENT_USAGE_ELEMENT = "document-usage";

    /**
     * Represents the attribute/element name where the string ID of the document index is stored.
     */
    private static final String DOCUMENT_ID_NODE = "document-id";

    /**
     * Represents the attribute/element name where the ID of the collection index is stored.
     */
    private static final String COLLECTION_ID_NODE = "collection-id";

    /**
     * Represents the element name where the collection index is stored.
     */
    private static final String COLLECTION_INDEX_ELEMENT = "document-collection";

    /**
     * Represents the attribute name where the document usage count is stored.
     */
    private static final String USAGE_COUNT_ATTR = "usage-count";

    /**
     * Represents the element name where the word index is stored.
     */
    private static final String WORD_INDEX_ELEMENT = "word-index";

    /**
     * Represents the element name where the word is stored.
     */
    private static final String WORD_ELEMENT = "word";

    /**
     * Represents the element name where the word position is stored.
     */
    private static final String POSITION_ELEMENT = "pos";

    /**
     * Represents the element name where the collection indices containing one document index is stored.
     */
    private static final String COLLECTION_INCLUSION_ELEMENT = "collection-inclusion";

    /**
     * Represents the element name where the ID of the collection index containing one document index is stored.
     */
    private static final String COLLECTION_ELEMENT = "collection";

    /**
     * Represents the length of a hex string for a unicode character.
     */
    private static final byte UNICODE_LENGTH = 4;

    /**
     * Represents the element name where the source instance of the document index is stored.
     */
    private static final String SOURCE_IDENTITY_ELEMENT = "source-identity";

    /**
     * Represents the element name where the class name of the document index is stored.
     */
    private static final String CLASS_NAME_ELEMENT = "class-name";

    /**
     * Represents the element name where the locale of the document index is stored.
     */
    private static final String LOCALE_ELEMENT = "locale";

    /**
     * Represents the element name where the delimiter of the document index is stored.
     */
    private static final String ENTRY_ELEMENT = "entry";

    /**
     * Represents the attribute name where the delimiter of the document index is stored.
     */
    private static final String DELIMITER_ATTRIBUTE = "delimiter";

    /**
     * Represents the attribute name where the locale language of the document index is stored.
     */
    private static final String LANGUAGE_ATTRIBUTE = "language";

    /**
     * Represents the attribute name where the locale country of the document index is stored.
     */
    private static final String COUNTRY_ATTRIBUTE = "country";

    /**
     * Represents the attribute name where the locale variant of the document index is stored.
     */
    private static final String VARIANT_ATTRIBUTE = "variant";

    /**
     * Represents the file name used to store the document usage count and collection IDs containing the document
     * indices.
     */
    private static final String DOC_USAGE_FILE_NAME = "document_usage.xml";

    /**
     * Represents the prefix of the file name used to store the collection indices.
     */
    private static final String DOC_COLLECTION_FILE_NAME_PREFIX = "doc_collection_";

    /**
     * Represents the prefix of the file name used to store the document indices.
     */
    private static final String DOC_INDEX_FILE_NAME_PREFIX = "doc_index_";

    /**
     * Represents the suffix of the first file used to store the document indices.
     */
    private static final String DOC_INDEX_FIRST_FILE_SUFFIX = "_1" + XML_EXTENSION;

    /**
     * Represents the regular expression pattern used to match all first document index files.
     */
    private static final Pattern FIRST_FILE_PATTERN = Pattern.compile(DOC_INDEX_FILE_NAME_PREFIX + "(.*)_1\\.xml");

    /**
     * Represents the radix of the hex integer.
     */
    private static final byte HEX_RADIX = 16;

    /**
     * Represents the buffer size used in BASE64 encoding.
     */
    private static final int BUFFER_SIZE = 1024;

    /**
     * Represent the file size margin. It is the maximum file size subtracting the file size safety margine. It is in
     * bytes. It might be a negative integer, which would cause each word storing in a separate file.
     */
    private final long fileSizeMargin;

    /**
     * Represents the base directory used to store all XML files. It must be an existing directory.
     */
    private final File persistenceDirPath;

    /**
     * Creates a new instance of <code>XmlIndexPersistence</code> class. The namespace used to configure this instance
     * is given.
     *
     * @param namespace the configuration namespace where all configuration properties are read.
     * @throws IllegalArgumentException if <code>namespace</code> is <code>null</code> or empty string.
     * @throws PersistenceConfigurationException if the configuration property is invalid.
     */
    public XmlIndexPersistence(String namespace) throws PersistenceConfigurationException {
        validateNotNullOrEmpty(namespace, "namespace");

        // Load configurations
        persistenceDirPath = new File(getConfigStringMandatory(namespace, XML_PATH_PROPERTY));
        fileSizeMargin = (getConfigIntegerMandatory(namespace, FILE_SIZE_LIMIT_PROPERTY, 1) * KILO
            - getConfigIntegerMandatory(namespace, FILE_SIZE_THRESHOLD_PROPERTY, 0)) * KILO;

        if (!persistenceDirPath.isDirectory()) {
            throw new PersistenceConfigurationException("persistenceDirPath must be an existing directory.");
        }
    }

    /**
     * Creates a new instance of <code>XmlIndexPersistence</code> class. The maximum file size, the file size safety
     * margine and the base directory of the XML files are given.
     *
     * @param fileSizeLimit the maximum file size of the document index files in mega-bytes (1024*1024 bytes).
     * @param fileSizeThreshold the file size safety margine of the document index files in kilo-bytes (1024 bytes).
     * @param persistenceDirPath the base directory where all XML files are stored.
     * @throws IllegalArgumentException if <code>fileSizeLimit</code> is non-positive; or
     *             <code>fileSizeThreshold</code> is negative; or <code>persistenceDirPath</code> is
     *             <code>null</code> or empty string; or <code>persistenceDirPath</code> is not an existing directory.
     */
    public XmlIndexPersistence(int fileSizeLimit, int fileSizeThreshold, String persistenceDirPath) {
        validateMinimum(fileSizeLimit, "fileSizeLimit", 1);
        validateMinimum(fileSizeThreshold, "fileSizeThreshold", 0);
        validateNotNullOrEmpty(persistenceDirPath, "persistenceDirPath");

        this.fileSizeMargin = (fileSizeLimit * KILO - fileSizeThreshold) * KILO;
        this.persistenceDirPath = new File(persistenceDirPath);

        if (!this.persistenceDirPath.isDirectory()) {
            throw new IllegalArgumentException("persistenceDirPath must be an existing directory.");
        }
    }

    /**
     * Gets the DOM document of the document usage XML file. When the document usage file does not exist, it fails. Note
     * that only the document element name is verified.
     *
     * @return the DOM document of the document usage file.
     * @throws IndexPersistenceException if loading the document usage file fails; or the document usage file corrupts.
     */
    private Document getDocumentUsage() throws IndexPersistenceException {
        File file = new File(persistenceDirPath, DOC_USAGE_FILE_NAME);

        try {
            Document doc = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(file);

            if (!doc.getDocumentElement().getNodeName().equals(DOCUMENTS_ELEMENT)) {
                // When the root element does not match, it is not a document usage file.
                throw new IndexPersistenceException("The XML file is not a document usage file.");
            }

            return doc;
        } catch (SAXException e) {
            throw new IndexPersistenceException("The XML file is not well-formed.", e);
        } catch (ParserConfigurationException e) {
            throw new IndexPersistenceException("XML parser cannot be created.", e);
        } catch (IOException e) {
            throw new IndexPersistenceException("Document usage XML cannot be read.", e);
        }
    }

    /**
     * Encodes the document source instance using the BASE64 encoding. The document source instance is serialized before
     * encoded.
     *
     * @param source the document source instance to be serialized and encoded.
     * @return the BASE64 encoded serialized document source instance.
     * @throws IndexPersistenceException if <code>source</code> cannot be serialized.
     */
    private static String encodeSource(Object source) throws IndexPersistenceException {
        // Use an object output stream wrapping a deflater output stream to encode
        Deflater deflater = new Base64Encoder(0, null, Base64Codec.STANDARD_ALPHABET);
        ByteArrayOutputStream baos = new ByteArrayOutputStream();

        try {
            ObjectOutputStream oos = new ObjectOutputStream(baos);

            try {
                // Serialize the source instance
                oos.writeObject(source);
            } finally {
                oos.close();
            }
        } catch (IOException e) {
            throw new IndexPersistenceException("The source instance cannot be serialized.", e);
        }

        // Encode
        deflater.setInput(baos.toByteArray());
        deflater.finish();

        // BASE64 could encode 1 byte into 4 bytes (using padding). Therefore, the current size * 2 is not enough
        byte[] buffer = new byte[BUFFER_SIZE];
        int length;

        baos.reset();

        // Repeat the encoding process until there is no more output from the encoder.
        while ((length = deflater.deflate(buffer)) != 0) {
            baos.write(buffer, 0, length);
        }

        try {
            // Convert the BASE64 encoded array to the UTF-8 encoded string.
            return baos.toString(ENCODING);
        } catch (UnsupportedEncodingException e) {
            // Almost impossible, since UTF-8 is widely used.
            throw new IndexPersistenceException("The UTF-8 encoding is not supported.", e);
        }
    }

    /**
     * Decodes the document source instance using the BASE64 encoding. The document source instance is deserialized
     * after decoded.
     *
     * @param data the encoded and serialized document source instance.
     * @return the decoded and deserialized document source instance.
     * @throws IndexPersistenceException if <code>data</code> cannot be decoded or deserialized.
     */
    private static Object decodeSource(String data) throws IndexPersistenceException {
        // Gets the bytes using UTF-8 encoding
        byte[] content;

        try {
            content = data.getBytes(ENCODING);
        } catch (UnsupportedEncodingException e) {
            // Almost impossible, since UTF-8 is widely used.
            throw new IndexPersistenceException("The UTF-8 encoding is not supported.", e);
        }

        // Use an object input stream wrapping a inflater input stream to decode
        Inflater inflater = new Base64Decoder(Base64Codec.IGNORE_NONE, null, Base64Codec.STANDARD_ALPHABET);
        int length;
        byte[] decoded = new byte[content.length];

        try {
            // Since decoding BASE64 always results a smaller content, it is safe here.
            inflater.setInput(content);
            length = inflater.inflate(decoded);
        } catch (DataFormatException e) {
            throw new IndexPersistenceException("The source is not BASE64 encoded.", e);
        }

        ByteArrayInputStream bais = new ByteArrayInputStream(decoded, 0, length);
        Object source;

        try {
            ObjectInputStream ois = new ObjectInputStream(bais);

            try {
                // Deserialize the source instance
                source = ois.readObject();
            } finally {
                ois.close();
            }
        } catch (IOException e) {
            throw new IndexPersistenceException("The source instance cannot be deserialized.", e);
        } catch (ClassNotFoundException e) {
            throw new IndexPersistenceException("The source instance class cannot be found.", e);
        }

        return source;
    }

    /**
     * Adds the given document index to the persistence. The initial collection reference count should be 0. When the
     * document index exists in the persistence, it fails. The document usage count entry for the document index is
     * created. The document index is written directly to the output streams. When the file is too large, it splits
     * automatically.
     *
     * @param documentIndex the document index to be persisted.
     * @throws IllegalArgumentException if <code>documentIndex</code> is <code>null</code>.
     * @throws IndexPersistenceException if persisting the document index fails; or the document index exists in the
     *             persistence; or the source identity instance is not serializable.
     */
    public void addDocumentIndex(DocumentIndex documentIndex) throws IndexPersistenceException {
        validateNotNull(documentIndex, "documentIndex");

        // Generate the string ID
        WordSourceId sourceId = documentIndex.getWordSourceId();
        Locale locale = sourceId.getSourceLocale();
        String id = IndexPersistenceHelper.generateId(sourceId);

        // Persist the document. The first file has sequence ID 1.
        int fileCount = 1;
        long byteCount = 0;
        File outputFile = new File(persistenceDirPath, DOC_INDEX_FILE_NAME_PREFIX + id + '_' + fileCount
            + XML_EXTENSION);

        if (outputFile.isFile()) {
            // When the first file exists, the document index is already persisted.
            throw new IndexPersistenceException("The document index is already in the persistence.");
        }

        // Encode the source identity.
        String encodedSource = encodeSource(sourceId.getSourceIdentity());
        StringBuffer buffer = new StringBuffer();
        byte[] header;

        // Build the necessary information.
        buffer.append(XML_HEADER);
        DOC_INDEX_XML_DOCUMENT_HEAD.format(new String[] {locale.toString(), id, encodedSource,
            sourceId.getSourceClassName(), locale.getLanguage(), locale.getCountry(), locale.getVariant()}, buffer,
            new FieldPosition(0));

        // Build the delimiters
        String[] delimiters = sourceId.getDelimiters();

        for (int delimiterIndex = 0; delimiterIndex < delimiters.length; ++delimiterIndex) {
            // Escape the delimiter characters
            buffer.append(DELIMITER_ENTRY_OPEN);
            writeEscaped(delimiters[delimiterIndex], buffer);
            buffer.append(DELIMITER_ENTRY_CLOSE);
        }

        buffer.append(DELIMITERS_END);
        OutputStream output = null;

        try {
            // Get the header bytes
            header = buffer.toString().getBytes(ENCODING);
            buffer.setLength(0);

            // Begin the actual writing
            byte[] content = null;
            Iterator iter = documentIndex.getWords().entrySet().iterator();

            do {
                // When it is the beginning of a new file, write the header information
                output = new BufferedOutputStream(new FileOutputStream(outputFile));
                output.write(header);
                byteCount = header.length;

                do {
                    // If there are data pending writing, write the data
                    if (content != null) {
                        output.write(content);
                        byteCount += content.length;
                        content = null;
                    }

                    if (!iter.hasNext()) {
                        // When there is no more word, no need to continue
                        break;
                    }

                    // Write the next word index
                    Entry entry = (Entry) iter.next();

                    // Write the word itself
                    buffer.append(WORD_INDEX_OPEN);
                    buffer.append(WORD_OPEN);
                    writeEscaped(((CollationKey) entry.getKey()).getSourceString(), buffer);
                    buffer.append(WORD_CLOSE);

                    // Write the positions
                    for (Iterator posIter = ((List) entry.getValue()).iterator(); posIter.hasNext();) {
                        buffer.append(POSITION_OPEN);
                        buffer.append(posIter.next());
                        buffer.append(POSITION_CLOSE);
                    }

                    buffer.append(WORD_INDEX_CLOSE);

                    // Extract the content to be written
                    content = buffer.toString().getBytes(ENCODING);
                    buffer.setLength(0);

                    // Continue writing in the same file if within the limit
                } while (content.length + byteCount < fileSizeMargin);

                // Either it reaches the file size limit, or there is no more word
                // The file should be closed properly
                output.write(DOC_INDEX_XML_DOCUMENT_END.getBytes(ENCODING));
                output.close();
                ++fileCount;
                outputFile = new File(persistenceDirPath,
                    DOC_INDEX_FILE_NAME_PREFIX + id + '_' + fileCount + XML_EXTENSION);
            } while (iter.hasNext());
        } catch (IOException e) {
            throw new IndexPersistenceException("Writing the document index file fails.", e);
        } finally {
            try {
                if (output != null) {
                    // Close the last file, even if error occurs.
                    output.close();
                }
            } catch (IOException e) {
                throw new IndexPersistenceException("Closing the document index file fails.", e);
            }
        }

        // Create the document usage count entry by setting it as 0
        setDocumentUseCount(id, 0);
    }

    /**
     * Gets the document index from the persistence. The ID of the document is given. When the document is not persisted
     * in the persistence, <code>null</code> should be returned. Initially, the document index file with sequence ID 1
     * is read. After reading file n, if file n+1 does not exist, it stops reading. Otherwise, continue to read file
     * n+1. For performance consideration, SAX parser is used to retrieve the document index data.
     *
     * @param wordSourceId the ID of the document whose index is retrieved.
     * @return the document index of the document with the ID.
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if retrieving the document index fails; or the source identity instance is not
     *             serializable.
     */
    public DocumentIndex getDocumentIndex(WordSourceId wordSourceId) throws IndexPersistenceException {
        String id = IndexPersistenceHelper.generateId(wordSourceId);

        // The first file sequence ID is 1
        int fileCount = 1;
        File file = new File(persistenceDirPath, DOC_INDEX_FILE_NAME_PREFIX + id + '_' + fileCount + XML_EXTENSION);

        if (!file.isFile()) {
            // When the first file cannot be found, the document index is not persisted.
            return null;
        }

        // Parse the XML file sets
        final Map wordMap = new HashMap();
        final Collator collator = Collator.getInstance(wordSourceId.getSourceLocale());

        collator.setStrength(Collator.PRIMARY);

        // While the XML file exists
        while (file.isFile()) {
            parseXml(file, new DefaultHandler() {
                /**
                 * Represents the state of expecting a word index.
                 */
                private static final byte EXPECT_WORD_INDEX = 0;

                /**
                 * Represents the state of expecting a word.
                 */
                private static final byte EXPECT_WORD = 1;

                /**
                 * Represents the state of expecting a position of a word.
                 */
                private static final byte EXPECT_POSITION = 2;

                /**
                 * Represents the list of positions of the current word being parsed. It may be <code>null</code>
                 * if there is not word being parsed.
                 */
                private List positions;

                /**
                 * Represents the word whose positions are parsed.
                 */
                private CollationKey word;

                /**
                 * Represents the buffer used to accumulate text sections encapulated by 'word' element or 'pos'
                 * element.
                 */
                private StringBuffer buffer = new StringBuffer();

                /**
                 * Represents a flag indicating whether the text sections should be accumulated.
                 */
                private boolean inText = false;

                /**
                 * Represents the current item being parsed.
                 */
                private byte state = EXPECT_WORD_INDEX;

                /**
                 * Processes the character data in the document. When the text is accumulated, the characters are
                 * added to the buffer.
                 *
                 * @param ch the characters to be processed.
                 * @param start the start position of the characters in the array.
                 * @param length the length of the characters in the array.
                 */
                public void characters(char[] ch, int start, int length) {
                    if (inText) {
                        // When the buffer is not null, accumulate the text
                        buffer.append(ch, start, length);
                    }
                }

                /**
                 * Processes the whitespaces in the document. When the text is accumulated, the whitespaces are
                 * added to the buffer.
                 *
                 * @param ch the whitespaces to be processed.
                 * @param start the start position of the whitespaces in the array.
                 * @param length the length of the whitespaces in the array.
                 */
                public void ignorableWhitespace(char[] ch, int start, int length) {
                    characters(ch, start, length);
                }

                /**
                 * Processes when the start of an element is found in the XML file.
                 *
                 * @param uri the namespace URI of the element.
                 * @param localName the local element tag name of the element.
                 * @param qName the full-qualified element name of the element.
                 * @param attributes the attributes of the element.
                 */
                public void startElement(String uri, String localName, String qName, Attributes attributes) {
                    if (WORD_INDEX_ELEMENT.equals(qName) && (state == EXPECT_WORD_INDEX)) {
                        // When the word index element is found, we expect the word.
                        state = EXPECT_WORD;

                        // Initialize the word and the positions
                        positions = new ArrayList();
                        word = null;
                    } else if ((WORD_ELEMENT.equals(qName) && (state == EXPECT_WORD))
                        || (POSITION_ELEMENT.equals(qName) && (state == EXPECT_POSITION))) {
                        // When the word element or position element is found, we start accumulating the text.
                        inText = true;
                    }
                }

                /**
                 * Processes when the end of an element is found in the XML file.
                 *
                 * @param uri the namespace URI of the element.
                 * @param localName the local element tag name of the element.
                 * @param qName the full-qualified element name of the element.
                 * @throws SAXException if the position of a word is not an integer; or the word string in XML file is
                 *             not properly escaped; or a word appears twice in the XML files.
                 */
                public void endElement(String uri, String localName, String qName) throws SAXException {
                    if (WORD_INDEX_ELEMENT.equals(qName) && (state == EXPECT_POSITION)) {
                        // When the word-index element is finished, we try to find the next word index element
                        state = EXPECT_WORD_INDEX;

                        // Add the word into the word map
                        if (wordMap.put(word, positions) != null) {
                            throw new SAXException("A word appears twice in the persistence.");
                        }
                    } else if (WORD_ELEMENT.equals(qName) && (state == EXPECT_WORD)) {
                        // When the word element is finished, set the word and clear the buffer.
                        // Since the word may contain some strange characters, it must be unescaped.
                        try {
                            word = collator.getCollationKey(unescape(buffer.toString()));
                        } catch (IllegalArgumentException e) {
                            throw new SAXException("The word must be properly escaped.", e);
                        }

                        buffer.setLength(0);
                        inText = false;

                        // Next, we expect position elements
                        state = EXPECT_POSITION;
                    } else if (POSITION_ELEMENT.equals(qName) && (state == EXPECT_POSITION)) {
                        // When the position element is finished, add the position to the current position list
                        try {
                            positions.add(Integer.valueOf(buffer.toString()));
                        } catch (NumberFormatException e) {
                            throw new SAXException("The position must be an integer.", e);
                        }

                        inText = false;
                        buffer.setLength(0);
                    }
                }
            });

            // Try to parse the next file.
            ++fileCount;
            file = new File(persistenceDirPath, DOC_INDEX_FILE_NAME_PREFIX + id + '_' + fileCount + XML_EXTENSION);
        }

        // See if all the document index files belonging to this document index are read.
        if (getDocumentIndexFiles(id).length != fileCount - 1) {
            // If not, some files are missing. It must be file corruption.
            throw new IndexPersistenceException("Some document index files are missing.");
        }

        try {
            return new DocumentIndex(wordSourceId, wordMap);
        } catch (IllegalArgumentException e) {
            throw new IndexPersistenceException("The persisted document index XML file is not valid.", e);
        }
    }

    /**
     * Gets all document index files belonging to this document string ID.
     *
     * @param id the string ID whose document index files are retrieved.
     * @return all document index files belonging to this document string ID.
     */
    private File[] getDocumentIndexFiles(String id) {
        final Pattern pattern = Pattern.compile(DOC_INDEX_FILE_NAME_PREFIX + id + DOC_INDEX_SEQUENCE_PATTERN,
            Pattern.CASE_INSENSITIVE);

        // Look for all files belonging to this document index
        return persistenceDirPath.listFiles(new FilenameFilter() {
            /**
             * Checks if the given file name should be contained in the result file array. All files belonging to the
             * document index to be removed are included.
             *
             * @param dir the directory of the file to be checked.
             * @param name the file name to be checked.
             * @return <code>true</code> if the file should be included in the result file array; <code>false</code>
             *         otherwise.
             */
            public boolean accept(File dir, String name) {
                return pattern.matcher(name).matches() && new File(dir, name).isFile();
            }
        });

    }

    /**
     * Removes a document index from the persistence. The document index can be deleted only when there is no more
     * collection containing the document index. The ID of the document is given. When the document is not persisted, it
     * fails. All document index files related to this document index are deleted.
     *
     * @param wordSourceId the ID of the document whose index is deleted.
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if deleting the document index fails; or the document index is not persisted;
     *             or there is at least one collection containing the document index; or the source identity instance is
     *             not serializable.
     */
    public void removeDocumentIndex(WordSourceId wordSourceId) throws IndexPersistenceException {
        if (getDocumentUseCount(wordSourceId) != 0) {
            // If the usage count is not 0, throw exception
            throw new IndexPersistenceException("The document index usage count is not 0.");
        }

        // Generate the regular expression pattern used to match files belonging to the document index
        String id = IndexPersistenceHelper.generateId(wordSourceId);
        File[] files = getDocumentIndexFiles(id);

        // Delete all found files
        for (int index = 0; index < files.length; ++index) {
            if (!files[index].delete()) {
                throw new IndexPersistenceException("Removing the document index from persistence fails.");
            }
        }

        // Remove the document usage count entry
        Document doc = getDocumentUsage();
        Element found = findChildElement(doc.getDocumentElement(), DOCUMENT_ID_NODE, id);

        if (found != null) {
            // If it is found, remove the element and persist
            doc.getDocumentElement().removeChild(found);
            saveDocumentUsageCountXml(doc);
        }
    }

    /**
     * Adds the collection index to the persistence. When the collection index exists in the persistence, it fails. The
     * collection index file is created. In addition, the collection inclusion part of the document usage file is
     * updated. It writes to the output stream directly.
     *
     * @param collectionIndex the collection index to be persisted.
     * @throws IllegalArgumentException if <code>collectionIndex</code> is <code>null</code>.
     * @throws IndexPersistenceException if persisting the collection index fails; or the collection index exists in the
     *             persistence; or the source identity instance of one document index is not serializable; or any
     *             document index in the collection is not persisted.
     */
    public void addCollectionIndex(CollectionIndex collectionIndex) throws IndexPersistenceException {
        validateNotNull(collectionIndex, "collectionIndex");

        // Check the existence of the collection index
        File file = new File(persistenceDirPath,
            DOC_COLLECTION_FILE_NAME_PREFIX + collectionIndex.getId() + XML_EXTENSION);

        if (file.isFile()) {
            throw new IndexPersistenceException("The collection index file exists.");
        }

        // Get the usage document
        Document doc = getDocumentUsage();

        // Call the actual function to add
        addCollectionIndexImpl(collectionIndex, doc);

        // Persist the updated document usage file
        saveDocumentUsageCountXml(doc);
    }

    /**
     * Adds the collection index to the persistence. When the collection index exists in the persistence, it fails. The
     * document usage DOM document is updated for the collection inclusion.
     *
     * @param collectionIndex the collection index to be persisted.
     * @param doc the document usage DOM document to be updated.
     * @throws IndexPersistenceException if persisting the collection index fails; or the collection index exists in the
     *             persistence; or the source identity instance of one document index is not serializable; or any
     *             document index in the collection is not persisted.
     */
    private void addCollectionIndexImpl(CollectionIndex collectionIndex, Document doc)
        throws IndexPersistenceException {
        String collectionId = collectionIndex.getId();
        File file = new File(persistenceDirPath, DOC_COLLECTION_FILE_NAME_PREFIX + collectionId + XML_EXTENSION);

        // Generate all string IDs of the document indices
        List allIds = new ArrayList();

        for (Iterator iter = collectionIndex.getAllDocumentIds().iterator(); iter.hasNext();) {
            String stringId = IndexPersistenceHelper.generateId((WordSourceId) iter.next());

            // Check if the document index is persisted
            if (!new File(persistenceDirPath,
                DOC_INDEX_FILE_NAME_PREFIX + stringId + DOC_INDEX_FIRST_FILE_SUFFIX).isFile()) {
                throw new IndexPersistenceException("One document index in the collection is not persisted.");
            }

            allIds.add(stringId);
        }

        // Write the collection index file.
        Writer writer = null;
        Element root = doc.getDocumentElement();

        try {
            writer = new OutputStreamWriter(new BufferedOutputStream(new FileOutputStream(file)), ENCODING);

            // Write the XML header
            writer.write(XML_HEADER);

            // Write the root element and the collection ID
            writeOpenTag(COLLECTION_INDEX_ELEMENT, Collections.singletonMap(COLLECTION_ID_NODE, collectionId), writer);
            writeOpenTag(DOCUMENTS_ELEMENT, Collections.EMPTY_MAP, writer);

            // Write all document string IDs in the collection index
            for (Iterator iter = allIds.iterator(); iter.hasNext();) {
                String stringId = (String) iter.next();

                // Write the document string ID to the collection file
                writeOpenTag(DOCUMENT_ID_NODE, Collections.EMPTY_MAP, writer);
                writer.write(stringId);
                writeCloseTag(DOCUMENT_ID_NODE, writer);

                // Add the collection ID to the document usage file
                Element inclusion = doc.createElement(COLLECTION_ELEMENT);

                inclusion.setAttribute(COLLECTION_ID_NODE, collectionId);
                findChildElement(root, DOCUMENT_ID_NODE, stringId).getFirstChild().appendChild(inclusion);
            }

            // Write the closing tags
            writeCloseTag(DOCUMENTS_ELEMENT, writer);
            writeCloseTag(COLLECTION_INDEX_ELEMENT, writer);
        } catch (NullPointerException e) {
            throw new IndexPersistenceException("The document index in the collection is not in usage document.", e);
        } catch (IOException e) {
            throw new IndexPersistenceException("Writing the collection index XML file fails.", e);
        } finally {
            if (writer != null) {
                try {
                    // Close the writer, even if error occurs.
                    writer.close();
                } catch (IOException e) {
                    throw new IndexPersistenceException("Closing the collection index XML file fails.", e);
                }
            }
        }
    }

    /**
     * Writes the closing tag to the writer. The tag name is given.
     *
     * @param tagName the tag name of the element to be written.
     * @param writer the writer where the closing tag is written to.
     * @throws IOException if error occurs when writing the closing tag to the writer.
     */
    private static void writeCloseTag(String tagName, Writer writer) throws IOException {
        writer.write('<');
        writer.write('/');
        writer.write(tagName);
        writer.write('>');
    }

    /**
     * Writes the openning tag to the writer. The tag name and the attributes of the element are given. The keys of the
     * attribute map are attribute names, and the values of the attribute map are attribute values.
     *
     * @param tagName the tag name of the element to be written.
     * @param attributes the attribute map of the element to be written.
     * @param writer the writer where the openning tag is written to.
     * @throws IOException if error occurs when writing the openning tag to the writer.
     */
    private static void writeOpenTag(String tagName, Map attributes, Writer writer) throws IOException {
        // Write the tag name
        writer.write('<');
        writer.write(tagName);

        // Write all attributes
        for (Iterator iter = attributes.entrySet().iterator(); iter.hasNext();) {
            Entry entry = (Entry) iter.next();

            // Write the attribute as key="value"
            writer.write(' ');
            writer.write(entry.getKey().toString());
            writer.write('=');
            writer.write('\"');
            writer.write(entry.getValue().toString());
            writer.write('\"');
        }

        writer.write('>');
    }

    /**
     * Gets the collection index from the persistence. The ID of the collection index is given. When the collection
     * index is not persisted, <code>null</code> should be returned.
     *
     * @param collectionId the ID of the collection index to be retrieved.
     * @return the collection index with the given ID.
     * @throws IllegalArgumentException if <code>collectionId</code> is <code>null</code> or empty string.
     * @throws IndexPersistenceException if retrieving the collection index fails.
     */
    public CollectionIndex getCollectionIndex(String collectionId) throws IndexPersistenceException {
        validateNotNullOrEmpty(collectionId, "collectionId");

        // Get all document string IDs in the collection
        Set stringIds = getStringIdsInCollection(collectionId);

        if (stringIds == null) {
            // When the collection index does not exist, return null
            return null;
        }

        // Build the word map and the document IDs
        Set sourceIds = new HashSet();
        Map wordMap = new HashMap();

        for (Iterator iter = stringIds.iterator(); iter.hasNext();) {
            // Read the word source ID first.
            WordSourceId wordSourceId = getWordSourceId(new File(persistenceDirPath, DOC_INDEX_FILE_NAME_PREFIX
                + iter.next() + DOC_INDEX_FIRST_FILE_SUFFIX));

            // Read the whole document index. There is very little overhead, since the word source ID part is small.
            DocumentIndex index = getDocumentIndex(wordSourceId);

            // Add the document
            sourceIds.add(wordSourceId);

            // Build the word map
            IndexPersistenceHelper.addDocumentWordMap(wordMap, index);
        }

        return new CollectionIndex(null, wordMap, sourceIds, collectionId);
    }

    /**
     * Removes the collection index from the persistence. The ID of the collection index is given. When the collection
     * index is not persisted, it fails. The collection index file is deleted. In addition, the collection inclusion
     * part of the document usage file is updated.
     *
     * @param collectionId the ID of the collection index to be deleted.
     * @throws IllegalArgumentException if <code>collectionId</code> is <code>null</code> or empty string.
     * @throws IndexPersistenceException if deleting the collection index fails; or the collection index ID cannot be
     *             found in the persistence.
     */
    public void removeCollectionIndex(String collectionId) throws IndexPersistenceException {
        validateNotNullOrEmpty(collectionId, "collectionId");

        // Get the usage document
        Document doc = getDocumentUsage();

        // Remove the collection from the usage document
        removeCollectionIndexImpl(collectionId, doc);

        // Finally, delete the collection index file
        if (!new File(persistenceDirPath, DOC_COLLECTION_FILE_NAME_PREFIX + collectionId + XML_EXTENSION).delete()) {
            throw new IndexPersistenceException("The collection index file cannot be deleted.");
        }

        // Persist the updated document usage file
        saveDocumentUsageCountXml(doc);
    }

    /**
     * Removes the collection index from the document usage DOM. The ID of the collection index is given. When the
     * collection index is not persisted, it fails. The document usage DOM document is also updated in memory.
     *
     * @param collectionId the ID of the collection index to be deleted.
     * @param doc the document usage DOM document to be updated.
     * @throws IndexPersistenceException if deleting the collection index fails; or the collection index ID cannot be
     *             found in the persistence.
     */
    private void removeCollectionIndexImpl(String collectionId, Document doc) throws IndexPersistenceException {
        // Get all document string IDs in the collection. Their collection inclusion should be modified.
        Set stringIds = getStringIdsInCollection(collectionId);

        if (stringIds == null) {
            // When the set is null, the collection index does not exist.
            throw new IndexPersistenceException("The collection index file does not exist.");
        }

        // Remove the collection ID from the document usage
        Element root = doc.getDocumentElement();

        try {
            for (Iterator iter = stringIds.iterator(); iter.hasNext();) {
                // First, find the element storing the document usage count
                Node inclusion = findChildElement(root, DOCUMENT_ID_NODE, (String) iter.next()).getFirstChild();

                // Then, find and remove the collection ID in the its collection inclusion element
                Node child = findChildElement(inclusion, COLLECTION_ID_NODE, collectionId);

                if (child == null) {
                    throw new IndexPersistenceException("The collection cannot be found in the collection inclusion.");
                }

                inclusion.removeChild(child);
            }
        } catch (NullPointerException e) {
            throw new IndexPersistenceException("The document usage file mismatches the collection index file.", e);
        }
    }

    /**
     * Gets the document ID from the given file. The given file should be a document index XML file.
     *
     * @param file the document index XML file from which a document ID is retrieved.
     * @return the document ID retrieved from the document index XML file.
     * @throws IndexPersistenceException if <code>file</code> is not a valid document index XML file.
     */
    private static WordSourceId getWordSourceId(File file) throws IndexPersistenceException {
        // Data extracted from the SAX parser.
        // Since the anonymous inner class requires final instances
        // they are declared as arrays in order to be modified.
        final String[] sourceData = new String[] {EMPTY};
        final String[] className = new String[] {null};
        final Locale[] locale = new Locale[] {null};
        final List delimiters = new ArrayList();

        // Parse the XML file
        parseXml(file, new DefaultHandler() {
            /**
             * Represents the buffer used to accumulate text sections encapulated by 'source-identity' element or
             * 'class-name' element.
             */
            private StringBuffer buffer = new StringBuffer();

            /**
             * Represents a flag indicating whether the text sections should be accumulated.
             */
            private boolean inText = false;

            /**
             * Processes the character data in the document. When the text is accumulated, the characters are added
             * to the buffer.
             *
             * @param ch the characters to be processed.
             * @param start the start position of the characters in the array.
             * @param length the length of the characters in the array.
             */
            public void characters(char[] ch, int start, int length) {
                if (inText) {
                    // When the buffer is not null, accumulate the text
                    buffer.append(ch, start, length);
                }
            }

            /**
             * Processes the whitespaces in the document. When the text is accumulated, the whitespaces are added to
             * the buffer.
             *
             * @param ch the whitespaces to be processed.
             * @param start the start position of the whitespaces in the array.
             * @param length the length of the whitespaces in the array.
             */
            public void ignorableWhitespace(char[] ch, int start, int length) {
                characters(ch, start, length);
            }

            /**
             * Processes when the start of an element is found in the XML file.
             *
             * @param uri the namespace URI of the element.
             * @param localName the local element tag name of the element.
             * @param qName the full-qualified element name of the element.
             * @param attributes the attributes of the element.
             * @throws StopParseException if the XML file reaches the word index part.
             * @throws SAXException if mandatory attributes in the XML file is missing; or the delimiter is not properly
             *             escaped.
             */
            public void startElement(String uri, String localName, String qName, Attributes attributes)
                throws SAXException {
                try {
                    if (SOURCE_IDENTITY_ELEMENT.equals(qName) || CLASS_NAME_ELEMENT.equals(qName)) {
                        // When 'source-identity' or 'class-name' begins, the text should be extracted
                        inText = true;
                    } else if (LOCALE_ELEMENT.equals(qName)) {
                        // When 'locale' begins, the locale attributes are extracted
                        locale[0] = new Locale(attributes.getValue(LANGUAGE_ATTRIBUTE),
                            attributes.getValue(COUNTRY_ATTRIBUTE), attributes.getValue(VARIANT_ATTRIBUTE));
                    } else if (ENTRY_ELEMENT.equals(qName)) {
                        // When 'entry' begins, the delimiter is extracted from the attribute
                        try {
                            delimiters.add(unescape(attributes.getValue(DELIMITER_ATTRIBUTE)));
                        } catch (IllegalArgumentException e) {
                            throw new SAXException("The delimiter must be properly escaped.", e);
                        }
                    } else if (WORD_INDEX_ELEMENT.equals(qName)) {
                        // When 'word-index' begins, stop parsing.
                        throw new StopParseException();
                    }
                } catch (NullPointerException e) {
                    throw new SAXException("A mandatory attribute is missing.", e);
                }
            }

            /**
             * Processes when the end of an element is found in the XML file.
             *
             * @param uri the namespace URI of the element.
             * @param localName the local element tag name of the element.
             * @param qName the full-qualified element name of the element.
             */
            public void endElement(String uri, String localName, String qName) {
                if (SOURCE_IDENTITY_ELEMENT.equals(qName)) {
                    // When 'source-identity' ends, the text should be extracted directly with no unescaping.
                    sourceData[0] = buffer.toString();
                    buffer.setLength(0);
                    inText = false;
                } else if (CLASS_NAME_ELEMENT.equals(qName)) {
                    // When 'class-name' ends, the text should be extracted directly with no unescaping.
                    className[0] = buffer.toString();
                    buffer.setLength(0);
                    inText = false;
                }
            }
        });

        try {
            // Create the document source ID.
            return new WordSourceId(decodeSource(sourceData[0]), className[0], (String[]) delimiters
                .toArray(new String[0]), locale[0]);
        } catch (IllegalArgumentException e) {
            throw new IndexPersistenceException("The persisted document source ID is invalid.", e);
        }
    }

    /**
     * Gets the document string IDs in a collection index. The ID of the collection index is given. When the collection
     * index does not exist in the persistence, <code>null</code> is returned.
     *
     * @param collectionId the ID of the collection index whose document string IDs are retrieved.
     * @return a set of document string IDs belonging to the collection index.
     * @throws IndexPersistenceException if the collection index file is invalid; or the collection index file cannot be
     *             read.
     */
    private Set getStringIdsInCollection(String collectionId) throws IndexPersistenceException {
        // Check the existence of the collection index
        File file = new File(persistenceDirPath, DOC_COLLECTION_FILE_NAME_PREFIX + collectionId + XML_EXTENSION);

        if (!file.isFile()) {
            // The collection index file does not exist.
            return null;
        }

        // Parse the collection index file using SAX parser
        final Set set = new HashSet();

        parseXml(file, new DefaultHandler() {
            /**
             * Represents the buffer used to accumulate text sections encapulated by 'document-id' element.
             */
            private StringBuffer buffer = new StringBuffer();

            /**
             * Represents a flag indicating whether the text sections should be accumulated.
             */
            private boolean inText = false;

            /**
             * Processes the character data in the document. When the text is accumulated, the characters are added to
             * the buffer.
             *
             * @param ch the characters to be processed.
             * @param start the start position of the characters in the array.
             * @param length the length of the characters in the array.
             */
            public void characters(char[] ch, int start, int length) {
                if (inText) {
                    // When the buffer is not null, accumulate the text
                    buffer.append(ch, start, length);
                }
            }

            /**
             * Processes when the start of an element is found in the XML file. When the element name is 'document-id',
             * its text value should be accumulated.
             *
             * @param uri the namespace URI of the element.
             * @param localName the local element tag name of the element.
             * @param qName the full-qualified element name of the element.
             * @param attributes the attributes of the element.
             */
            public void startElement(String uri, String localName, String qName, Attributes attributes) {
                if (DOCUMENT_ID_NODE.equals(qName)) {
                    inText = true;
                }
            }

            /**
             * Processes when the end of an element is found in the XML file. When the element name is 'document-id',
             * its text value should be added to the set as the document string ID.
             *
             * @param uri the namespace URI of the element.
             * @param localName the local element tag name of the element.
             * @param qName the full-qualified element name of the element.
             */
            public void endElement(String uri, String localName, String qName) {
                if (DOCUMENT_ID_NODE.equals(qName)) {
                    inText = false;
                    set.add(buffer.toString());
                    buffer.setLength(0);
                }
            }
        });

        return set;
    }

    /**
     * Updates the collection index in the persistence. When the collection index ID cannot be found in the persistence,
     * it fails. The collection index is simply removed and re-added.
     *
     * @param collectionIndex the collection index to be updated.
     * @throws IllegalArgumentException if <code>collectionIndex</code> is <code>null</code>.
     * @throws IndexPersistenceException if updating the collection index fails; or the collection index ID cannot be
     *             found in the persistence; or the source identity instance of one document index is not serializable;
     *             or any document index in the collection is not persisted.
     */
    public void updateCollectionIndex(CollectionIndex collectionIndex) throws IndexPersistenceException {
        validateNotNull(collectionIndex, "collectionIndex");

        // Get the usage document
        Document doc = getDocumentUsage();

        // Update is the same process as first remove and then add.
        // The only overhead is updating the DOM document in memory which is pretty trivial.
        removeCollectionIndexImpl(collectionIndex.getId(), doc);
        addCollectionIndexImpl(collectionIndex, doc);

        // Persist the updated document usage file
        saveDocumentUsageCountXml(doc);
    }

    /**
     * Increases the collection reference count of the document index. The document ID is given. When the document index
     * is not persisted, it fails. The document usage file is updated.
     *
     * @param wordSourceId the ID of the document whose collection reference count is increased by 1.
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if increasing the collection reference count fails; or the document index is
     *             not persisted; or the source identity instance is not serializable.
     */
    public void increaseDocumentUseCount(WordSourceId wordSourceId) throws IndexPersistenceException {
        setDocumentUseCount(IndexPersistenceHelper.generateId(wordSourceId), getDocumentUseCount(wordSourceId) + 1);
    }

    /**
     * Decreases the collection reference count of the document index. The document ID is given. When the document index
     * is not persisted, it fails. The document usage file is updated.
     *
     * @param wordSourceId the ID of the document whose collection reference count is decreased by 1.
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if decreasing the collection reference count fails; or the document index is
     *             not persisted; or the collection reference count is already 0; or the source identity instance is not
     *             serializable.
     */
    public void decreaseDocumentUseCount(WordSourceId wordSourceId) throws IndexPersistenceException {
        int count = getDocumentUseCount(wordSourceId);

        if (count == 0) {
            // When the count is already 0, error
            throw new IndexPersistenceException("The document usage count is already 0.");
        }

        setDocumentUseCount(IndexPersistenceHelper.generateId(wordSourceId), count - 1);
    }

    /**
     * Gets a set of all IDs of documents whose indices are persisted in the persistence. All first document index files
     * are read.
     *
     * @return a set of all IDs of persisted document indices.
     * @throws IndexPersistenceException if retrieving the document IDs fails.
     */
    public Set getIndexedDocuments() throws IndexPersistenceException {
        Set set = new HashSet();

        // List all first document index files.
        File[] files = persistenceDirPath.listFiles(new FilenameFilter() {
            /**
             * Checks if the given file name should be contained in the result file array. All first document index
             * files are included.
             *
             * @param dir the directory of the file to be checked.
             * @param name the file name to be checked.
             * @return <code>true</code> if the file should be included in the result file array; <code>false</code>
             *         otherwise.
             */
            public boolean accept(File dir, String name) {
                return FIRST_FILE_PATTERN.matcher(name).matches() && new File(dir, name).isFile();
            }
        });

        // Get all document source IDs from the files
        for (int index = 0; index < files.length; ++index) {
            set.add(getWordSourceId(files[index]));
        }

        return set;
    }

    /**
     * Gets the collection reference count of the given document index. When the document index is not persisted, it
     * fails. SAX parser is used due to performance consideration.
     *
     * @param wordSourceId the ID of the document index whose collection reference count is retrieved.
     * @return the reference count of the document index.
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if retrieving the collection reference count fails; or the document index is
     *             not persisted; or the source identity instance is not serializable.
     */
    public int getDocumentUseCount(WordSourceId wordSourceId) throws IndexPersistenceException {
        final String id = IndexPersistenceHelper.generateId(wordSourceId);

        // Since the anonymous inner class requires a final variable, an array is used.
        final int[] count = new int[] {-1};

        // Parse the XML file.
        parseXml(new File(persistenceDirPath, DOC_USAGE_FILE_NAME), new DefaultHandler() {
            /**
             * Processes when the start of an element is found in the XML file. When the element is a
             * 'document_usage' element, and its document string ID is the expected ID, the document usage count
             * attribute is parsed and saved.
             *
             * @param uri the namespace URI of the element.
             * @param localName the local element tag name of the element.
             * @param qName the full-qualified element name of the element.
             * @param attributes the attributes of the element.
             * @throws StopParseException if the document usage count is found.
             * @throws SAXException if the document usage count attribute value is not an integer.
             */
            public void startElement(String uri, String localName, String qName, Attributes attributes)
                throws SAXException {
                if (DOCUMENT_USAGE_ELEMENT.equals(qName) && id.equals(attributes.getValue(DOCUMENT_ID_NODE))) {
                    try {
                        count[0] = Integer.parseInt(attributes.getValue(USAGE_COUNT_ATTR));
                    } catch (NumberFormatException e) {
                        throw new SAXException("The document usage count must be an integer.", e);
                    }

                    // The document is found, no need to parse more
                    throw new StopParseException();
                }
            }
        });

        if (count[0] < 0) {
            // When the count is negative, either the document index is not found, or the usage count is negative.
            throw new IndexPersistenceException("The document usage count cannot found, or is a negative value.");
        }

        return count[0];
    }

    /**
     * Parses an XML file using SAX parser. The handler used to process the XML file is given.
     *
     * @param file the XML file to be parsed.
     * @param handler the SAX handler used to process the XML file.
     * @throws IndexPersistenceException if a new SAX parser cannot be created; or the XML file cannot be parsed or
     *             read.
     */
    private static void parseXml(File file, DefaultHandler handler) throws IndexPersistenceException {
        InputStream input = null;

        try {
            // Create an input stream. This is to ensure that the file is closed even if error occurs.
            input = new BufferedInputStream(new FileInputStream(file));

            SAXParserFactory.newInstance().newSAXParser().parse(input, handler);
        } catch (StopParseException e) {
            // Stop parsing rest of the XML file.
        } catch (IOException e) {
            throw new IndexPersistenceException("Reading the XML file fails.", e);
        } catch (ParserConfigurationException e) {
            throw new IndexPersistenceException("SAX parser cannot be created.", e);
        } catch (SAXException e) {
            throw new IndexPersistenceException("SAX parser cannot parse the XML file successfully.", e);
        } finally {
            if (input != null) {
                try {
                    input.close();
                } catch (IOException e) {
                    throw new IndexPersistenceException("Closing the XML file fails.", e);
                }
            }
        }
    }

    /**
     * Sets the document use count for one document index. The string ID of the document index is given. The document
     * usage count XML file is updated. When the document index cannot be found in the XML file, a new entry is created.
     *
     * @param id the string ID fo the document index whose usage count is updated.
     * @param count the usage count of the document index to be updated.
     * @throws IndexPersistenceException if error occurs when reading or persisting the document usage count XML file.
     */
    private void setDocumentUseCount(String id, int count) throws IndexPersistenceException {
        // Get the element
        Document doc = getDocumentUsage();

        // Look for the document index entry
        Element found = findChildElement(doc.getDocumentElement(), DOCUMENT_ID_NODE, id);

        if (found == null) {
            // When the proper element is not found, create one
            found = doc.createElement(DOCUMENT_USAGE_ELEMENT);
            found.setAttribute(DOCUMENT_ID_NODE, id);
            found.appendChild(doc.createElement(COLLECTION_INCLUSION_ELEMENT));
            doc.getDocumentElement().appendChild(found);
        }

        // Set the count attribute
        found.setAttribute(USAGE_COUNT_ATTR, Integer.toString(count));

        // Persist it
        saveDocumentUsageCountXml(doc);
    }

    /**
     * Finds a child element directly nested under the given node. The given attribute of the child element must be the
     * given value. The first matched child is returned. If there is no such child, <code>null</code> is returned.
     *
     * @param root the parent node whose direct children are searched.
     * @param attribute the attribute name of the element to be looked for.
     * @param value the value of the given attribute of the element to be looked for.
     * @return the first direct child element whose attribute value matches.
     * @throws NullPointerException if <code>root</code> is <code>null</code>.
     */
    private static Element findChildElement(Node root, String attribute, String value) {
        NodeList list = root.getChildNodes();

        // Enumerate every child
        for (int index = 0; index < list.getLength(); ++index) {
            Node node = list.item(index);

            if ((node.getNodeType() == Node.ELEMENT_NODE) && value.equals(((Element) node).getAttribute(attribute))) {
                // Only child elements are considered. If the attribute value matches, we found it.
                return (Element) node;
            }
        }

        // Otherwise, it is not found.
        return null;
    }

    /**
     * Saves the DOM document as the document usage count XML file. UTF-8 encoding is used to save the XML file.
     *
     * @param doc the DOM document to be saved.
     * @throws IndexPersistenceException if error occurs when saving the DOM document as the XML file.
     */
    private void saveDocumentUsageCountXml(Document doc) throws IndexPersistenceException {
        Writer writer = null;

        try {
            // Create a UTF-8 writer and transform the DOM document into the writer
            writer = new OutputStreamWriter(new BufferedOutputStream(new FileOutputStream(new File(persistenceDirPath,
                DOC_USAGE_FILE_NAME))), ENCODING);
            TransformerFactory.newInstance().newTransformer().transform(new DOMSource(doc), new StreamResult(writer));
        } catch (IOException e) {
            throw new IndexPersistenceException("Writing the document usage count XML file fails.", e);
        } catch (TransformerConfigurationException e) {
            throw new IndexPersistenceException("XSLT transformer cannot be created.", e);
        } catch (TransformerException e) {
            throw new IndexPersistenceException("The DOM document cannot be transformed into XML file.", e);
        } finally {
            if (writer != null) {
                try {
                    writer.close();
                } catch (IOException e) {
                    throw new IndexPersistenceException("Closing the XML writer fails.", e);
                }
            }
        }
    }

    /**
     * Validates the value of a variable with type <code>Object</code>. The value cannot be <code>null</code>.
     *
     * @param value value of the variable.
     * @param name name of the variable.
     * @throws IllegalArgumentException if <code>value</code> is <code>null</code>.
     */
    private static void validateNotNull(Object value, String name) {
        if (value == null) {
            throw new IllegalArgumentException(name + " cannot be null.");
        }
    }

    /**
     * Validates the value of a string variable. The value cannot be <code>null</code> and empty string.
     *
     * @param value value of the variable.
     * @param name name of the variable.
     * @throws IllegalArgumentException if <code>value</code> is <code>null</code> or empty string.
     */
    private static void validateNotNullOrEmpty(String value, String name) {
        validateNotNull(value, name);

        if (value.trim().length() == 0) {
            throw new IllegalArgumentException(name + " cannot be empty string.");
        }
    }

    /**
     * Validates the value of an integer variable. The value cannot be smaller than the given minimum value.
     *
     * @param value value of the variable.
     * @param name name of the variable.
     * @param minimum minimum allowed value.
     * @throws IllegalArgumentException if <code>value</code> is smaller than <code>minimum</code>.
     */
    private static void validateMinimum(int value, String name, int minimum) {
        if (value < minimum) {
            throw new IllegalArgumentException(name + " cannot be smaller than " + minimum);
        }
    }

    /**
     * Gets the property value in the configuration manager. The property value should be an integer not smaller than
     * the given minimum value. The property must exist in the configuration.
     *
     * @param namespace the namespace where the property exists.
     * @param propertyName the property name whose value is retrieved.
     * @param minimum the minimum allowed value of the property.
     * @return the property value loaded from the configuration.
     * @throws PersistenceConfigurationException if the property is missing; or the property value is not an integer; or
     *             the property value is smaller than <code>minimum</code>; or accessing the configuration manager
     *             fails.
     */
    private static int getConfigIntegerMandatory(String namespace, String propertyName, int minimum)
        throws PersistenceConfigurationException {
        int value;

        try {
            // Get the property value from the configuration
            value = Integer.parseInt(getConfigStringMandatory(namespace, propertyName));
        } catch (NumberFormatException e) {
            // When the property value is not an integer.
            throw new PersistenceConfigurationException("The property '" + propertyName + "' is not an integer.");
        }

        if (value < minimum) {
            // When the value is smaller than the minimum value
            throw new PersistenceConfigurationException("The property value is smaller than the minimum value of "
                + minimum + ".");
        }

        return value;
    }

    /**
     * Gets the property value in the configuration manager. The property must exist in the configuration. Its value
     * cannot be empty string.
     *
     * @param namespace the namespace where the property exists.
     * @param propertyName the property name whose value is retrieved.
     * @return the property value loaded from the configuration.
     * @throws PersistenceConfigurationException if the property is missing; or the property value is empty string; or
     *             accessing the configuration manager fails.
     */
    private static String getConfigStringMandatory(String namespace, String propertyName)
        throws PersistenceConfigurationException {
        String value;

        try {
            // Load the value as string
            value = ConfigManager.getInstance().getString(namespace, propertyName);
        } catch (ConfigManagerException e) {
            throw new PersistenceConfigurationException("The configuration value cannot be read.", e);
        }

        // When the property is missing, throw exception
        if (value == null) {
            throw new PersistenceConfigurationException("The mandatory property '" + propertyName + "' is missing.");
        }

        // When the property is empty string, throw exception
        if (value.trim().length() == 0) {
            throw new PersistenceConfigurationException("The property '" + propertyName + "' is empty string.");
        }

        return value;
    }

    /**
     * Writes the source string to the XML file. The source string is properly escaped when necessary. It is used to
     * write the attribute value and text value.
     *
     * @param source the string to be written.
     * @param buffer the buffer where the escaped string is written to.
     */
    private static void writeEscaped(String source, StringBuffer buffer) {
        int length = source.length();

        for (int charIndex = 0; charIndex < length; ++charIndex) {
            char ch = source.charAt(charIndex);

            if (isXmlEscapeCharacter(ch)) {
                // When the character is a character need to be escaped
                buffer.append(ESCAPE_CHARACTER_PREFIX);
                buffer.append((int) ch);
                buffer.append(';');
            } else if (!isXmlCharacter(ch)) {
                // When the character is a character not allowed in the XML, escape it using \\uXXXX
                buffer.append(ESCAPE_CHARACTER);
                buffer.append('u');

                // Convert the character to a four-digit hex string
                String hex = Integer.toHexString(ch);

                for (int i = hex.length(); i < UNICODE_LENGTH; ++i) {
                    buffer.append('0');
                }

                buffer.append(hex);
            } else {
                buffer.append(ch);
            }
        }
    }

    /**
     * Unescapes the escaped string. The escaped string may contain only the form '&#92;uXXXX'. In this case, such
     * substring is converted into one unicode character, as the same in Java language.
     *
     * @param source the escaped string to be unescaped.
     * @return the unescaped string created from the source.
     * @throws NullPointerException if <code>source</code> is <code>null</code>.
     * @throws IllegalArgumentException if <code>source</code> is not a properly escaped string.
     */
    private static String unescape(String source) {
        StringBuffer buffer = new StringBuffer();
        int pos = 0;

        // Process every character in the source string
        while (pos < source.length()) {
            if (source.charAt(pos) == ESCAPE_CHARACTER) {
                // When the character is '&#92;', the following must be 'uXXXX'.
                try {
                    if (source.charAt(++pos) != 'u') {
                        // Illegal escape
                        throw new IllegalArgumentException("The escaped string is illegal.");
                    }

                    // Parse the unicode value and convert it into the character
                    ++pos;
                    buffer.append((char) Integer.parseInt(source.substring(pos, pos + UNICODE_LENGTH), HEX_RADIX));
                    pos += UNICODE_LENGTH;
                } catch (IndexOutOfBoundsException e) {
                    throw new IllegalArgumentException("The escaped string is illegal.");
                }
            } else {
                buffer.append(source.charAt(pos++));
            }
        }

        return buffer.toString();
    }

    /**
     * Gets a flag indicating whether a character is allowed in an XML file. See XML 1.0 specification. In addition,
     * '&#92;' is considered as invalid, since it is used as the escaping character.
     *
     * @param ch the character to be checked.
     * @return <code>true</code> if the character is allowed in the XML file; <code>false</code> otherwise.
     */
    private static boolean isXmlCharacter(char ch) {
        return (ch == '\t') || (ch == '\n') || (ch == '\r')
            || !((ch == ESCAPE_CHARACTER) || (ch == '\uFFFE') || (ch == '\uFFFF') || Character.isISOControl(ch)
                || (Character.getType(ch) == Character.SURROGATE));
    }

    /**
     * Gets a flag indicating whether a character needs to be escaped using XML escape rule. Tab, CR, LF, '&', '"', '<'
     * and '>' need to be escaped.
     *
     * @param ch the character to be checked.
     * @return <code>true</code> if the character is allowed in the XML file; <code>false</code> otherwise.
     */
    private static boolean isXmlEscapeCharacter(char ch) {
        return (ch == '&') || (ch == '"') || (ch == '<') || (ch == '>') || (ch == '\t') || (ch == '\n') || (ch == '\r');
    }

    /**
     * Defines a customized exception which is used to stop parsing the XML file. It is useful when the desired
     * information is acquired, but the rest of the XML file is huge.
     *
     * @author TCSDEVELOPER
     * @version 1.0
     */
    private static class StopParseException extends SAXException {
        /**
         * Creates a new instance of <code>StopParseException</code> class.
         */
        public StopParseException() {
            super((String) null);
        }
    }
}







