/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence;

import java.io.File;
import java.util.Iterator;
import java.util.Locale;
import java.util.Set;

import com.topcoder.document.index.CollectionIndex;
import com.topcoder.document.index.DocumentIndex;
import com.topcoder.document.index.IndexManager;
import com.topcoder.document.index.persistence.impl.xml.XmlIndexPersistence;
import com.topcoder.document.index.wordsource.WordSource;
import com.topcoder.document.index.wordsource.WordSourceId;
import com.topcoder.document.index.wordsource.textfile.TextFileWordSource;
import com.topcoder.util.config.ConfigManager;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

/**
 * Demonstrates the usage of this component. This component can be used in two ways, either as a plugin to the Document
 * Indexer 2.0 component, or as an independent component. Since the three classes provided by this component have
 * exactly the same interface and the similar behavior, only one of them is used to demonstrate. Since this is a
 * demonstration, there is no assertion in the test methods.
 *
 * @author TCSDEVELOPER
 * @version 1.0
 */
public class Demo extends TestCase {
    /**
     * Represents the configuration manager used in tests.
     */
    private static final ConfigManager CONFIG = ConfigManager.getInstance();

    /**
     * Represents the persistence base directory used in tests.
     */
    private static final File BASE_DIR = new File("test_files/Xml");

    /**
     * Aggragates all tests in this class.
     *
     * @return test suite aggragating all tests.
     */
    public static Test suite() {
        return new TestSuite(Demo.class);
    }

    /**
     * Deletes all XML files in the persistence directory.
     */
    private static void deleteXmlFiles() {
        // Delete all files in the persistence directory.
        File[] files = BASE_DIR.listFiles();

        for (int i = 0; i < files.length; ++i) {
            files[i].delete();
        }
    }

    /**
     * Sets up the test environment. All previous configurations are removed. The valid configuration for XML
     * persistence is loaded. The persistence directory is prepared.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    protected void setUp() throws Exception {
        for (Iterator iter = CONFIG.getAllNamespaces(); iter.hasNext();) {
            CONFIG.removeNamespace((String) iter.next());
        }

        CONFIG.add("XmlConfig/Valid.xml");

        deleteXmlFiles();
        TestHelper.copy(new File("test_files/XmlReference/document_usage_valid.xml"), new File(BASE_DIR,
            "document_usage.xml"));
    }

    /**
     * Cleans up the test environment. All previous configurations are removed. The persistence directory is cleaned.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    protected void tearDown() throws Exception {
        for (Iterator iter = CONFIG.getAllNamespaces(); iter.hasNext();) {
            CONFIG.removeNamespace((String) iter.next());
        }

        deleteXmlFiles();
    }

    /**
     * Demonstrates the usage of this component as a plugin to the Document Indexer 2.0 component. It is used together
     * with the <code>IndexManager</code> class in the Document Indexer 2.0 component.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testDemoIndexManager() throws Exception {
        // Create the index persistence by configuration
        IndexPersistence persistence = new XmlIndexPersistence("Test");

        // Create an index manager
        IndexManager manager = new IndexManager(persistence);

        // Manipulate the index manager. Please refer to the Document Indexer 2.0 component
        // Create two documents with default locale English (US).
        WordSource doc1 = new TextFileWordSource("test_files/Press.txt", TextFileWordSource.DEFAULT_DELIMITERS);
        WordSource doc2 = new TextFileWordSource("test_files/Press2.txt", TextFileWordSource.DEFAULT_DELIMITERS);

        // Create and persist the document indices via index manager
        DocumentIndex docIndex1 = manager.createDocumentIndex(doc1);
        DocumentIndex docIndex2 = manager.createDocumentIndex(doc2);

        // Create a collection index containing the two document indices
        CollectionIndex collection = manager.createCollectionIndex("collection");

        collection.addDocument(doc1);
        collection.addDocument(doc2);

        // Get the use count of the document
        System.out.println("Press1 is used by " + manager.getDocumentUseCount(docIndex1.getWordSourceId())
            + " collection(s).");
        System.out.println("Press2 is used by " + manager.getDocumentUseCount(docIndex2.getWordSourceId())
            + " collection(s).");

        // Remove the collection index
        collection.removeDocument(docIndex1.getWordSourceId());
        collection.removeDocument(docIndex2.getWordSourceId());
        manager.removeCollectionIndex("collection");

        // Remove the document indices
        manager.removeDocumentIndex(docIndex1.getWordSourceId());
        manager.removeDocumentIndex(docIndex2.getWordSourceId());
    }

    /**
     * Demonstrates the usage of component as an independent component. Though <code>IndexManager</code> class is not
     * used, it still needs Document Indexer 2.0 component. In this case, operations on collection index should not be
     * done, since they require the support of <code>IndexManager</code>.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testDemoIndependent() throws Exception {
        // Create a persistence by program.
        IndexPersistence persistence = new XmlIndexPersistence(10, 5, "test_files/Xml");

        // Create two documents with locale English (UK).
        WordSource doc1 = new TextFileWordSource("test_files/Press3.txt", TextFileWordSource.DEFAULT_DELIMITERS,
            Locale.UK);
        WordSource doc2 = new TextFileWordSource("test_files/Press4.txt", TextFileWordSource.DEFAULT_DELIMITERS,
            Locale.UK);

        // Persist the document indices of the two documents
        persistence.addDocumentIndex(new DocumentIndex(doc1));
        persistence.addDocumentIndex(new DocumentIndex(doc2));

        // Retrieve the persisted document index
        DocumentIndex index = persistence.getDocumentIndex(doc1.getId());

        // Increase the use count of the document index
        persistence.increaseDocumentUseCount(index.getWordSourceId());

        // Get the use count of the document index
        System.out.println("The use count of Press3 is " + persistence.getDocumentUseCount(index.getWordSourceId()));

        // Decrease the use count of the document index
        persistence.decreaseDocumentUseCount(index.getWordSourceId());

        // Get all persisted document IDs
        Set documentIds = persistence.getIndexedDocuments();

        // Delete all persisted document indices
        for (Iterator iter = documentIds.iterator(); iter.hasNext();) {
            persistence.removeDocumentIndex((WordSourceId) iter.next());
        }
    }
}








