/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.db;

import java.sql.Blob;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.text.Collator;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import com.topcoder.db.connectionfactory.DBConnectionFactory;
import com.topcoder.db.connectionfactory.DBConnectionFactoryImpl;
import com.topcoder.document.index.CollectionIndex;
import com.topcoder.document.index.DocumentIndex;
import com.topcoder.document.index.persistence.IndexPersistenceException;
import com.topcoder.document.index.persistence.TestHelper;
import com.topcoder.document.index.persistence.impl.IndexPersistenceHelper;
import com.topcoder.document.index.wordsource.WordSourceId;
import com.topcoder.util.config.ConfigManager;

import junit.framework.Test;
import junit.framework.TestSuite;

/**
 * Tests functionality and error cases of <code>SimpleAccessDBIndexPersistence</code> class.
 *
 * @author TCSDEVELOPER
 * @version 1.0
 */
public class SimpleAccessDBIndexPersistenceTestCase extends CommonDBIndexPersistenceTestCase {
    /**
     * Represents the configuration manager used in tests.
     */
    private static final ConfigManager CONFIG = ConfigManager.getInstance();

    /**
     * Represents the DB connection factory used in tests.
     */
    private DBConnectionFactory factory;

    /**
     * Aggragates all tests in this class.
     *
     * @return test suite aggragating all tests.
     */
    public static Test suite() {
        return new TestSuite(SimpleAccessDBIndexPersistenceTestCase.class);
    }

    /**
     * Creates a new instance of the DB index persistence used in tests.
     *
     * @return a new instance of the DB index persistence to be tested.
     * @throws Exception if error occurs when creating the index persistence.
     */
    protected AbstractDBIndexPersistence createPersistence() throws Exception {
        for (Iterator iter = CONFIG.getAllNamespaces(); iter.hasNext();) {
            CONFIG.removeNamespace((String) iter.next());
        }

        CONFIG.add("DBFactory.xml");
        factory = new DBConnectionFactoryImpl("com.topcoder.db.connectionfactory.DBConnectionFactoryImpl");

        for (Iterator iter = CONFIG.getAllNamespaces(); iter.hasNext();) {
            CONFIG.removeNamespace((String) iter.next());
        }

        return new SimpleAccessDBIndexPersistence(factory, null);
    }

    /**
     * Tests <code>getCollectionIndex(String)</code> when the document index in the collection is illegal in the
     * persistence. Expect IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetCollectionIndexStringIllegalDocumentIndex() throws Exception {
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US);
        String stringId = IndexPersistenceHelper.generateId(id);

        // Add a document index with no words
        persistence.addDocumentIndex(new DocumentIndex(id, Collections.EMPTY_MAP));

        // Add a collection index
        persistence
            .addCollectionIndex(new CollectionIndex(null, Collections.EMPTY_MAP, Collections.singleton(id), "id"));

        // Update the document index, so that a word has negative position
        DBPersistenceHelper.executeSQL(connection,
            "INSERT INTO WORD (DOCUMENT_ID, WORD_ID, WORD_ENTRY) VALUES (?, ?, ?)", new Object[] {stringId,
                new Integer(1), "word"});
        DBPersistenceHelper.executeSQL(connection,
            "INSERT INTO WORD_POSITION (DOCUMENT_ID, WORD_ID, WORD_POSITION)" + "VALUES (?, ?, ?)",
            new Object[] {stringId, new Integer(1), new Integer(-1)}).close();

        try {
            persistence.getCollectionIndex("id");
            fail("The document index file is invalid, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>addDocumentIndex(DocumentIndex)</code>. A new document record should be created in
     * table DOCUMENT. Two delimiter records should be created in table DELIMITER. The word records should be created in
     * table WORD. The word position records should be created in the table WORD_POSITION. These are verified by SQL
     * statement.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAddDocumentIndexDocumentIndexAccuracy() throws Exception {
        // Create the word map
        Map wordMap = new HashMap();
        Collator collator = Collator.getInstance(Locale.US);

        collator.setStrength(Collator.PRIMARY);

        for (int i = 0; i < 6; ++i) {
            List positions = new ArrayList();

            for (int j = 0; j < 10; ++j) {
                positions.add(new Integer(j));
            }

            wordMap.put(collator.getCollationKey("Word" + Integer.toString(i)), positions);

            // Build the word data
            StringBuffer buffer = new StringBuffer("Word" + Integer.toString(i));

            for (Iterator iter = positions.iterator(); iter.hasNext();) {
                buffer.append(' ');
                buffer.append(iter.next());
            }
        }

        // Create the document index
        WordSourceId sourceId = new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US);
        String stringId = IndexPersistenceHelper.generateId(sourceId);
        DocumentIndex index = new DocumentIndex(sourceId, wordMap);

        // Persist it
        persistence.addDocumentIndex(index);

        // Verify the document record
        PreparedStatement statement = DBPersistenceHelper.executeSQL(connection,
            "SELECT * FROM DOCUMENT WHERE DOCUMENT_ID = ?", new Object[] {stringId});

        try {
            ResultSet rs = statement.getResultSet();

            rs.next();

            Blob blob = rs.getBlob("SOURCE_IDENTITY");

            assertEquals("The document locale should be persisted.", Locale.US.toString(), rs.getString("LOCALE"));
            assertEquals("The document class name should be persisted.", "class", rs.getString("CLASS_NAME"));
            assertEquals("The document use count should be 0.", 0, rs.getInt("USE_COUNT"));
            assertTrue("The document source identity should be persisted.", Arrays.equals(TestHelper.serialize("id"),
                blob.getBytes(1, (int) blob.length())));
        } finally {
            statement.close();
        }

        // Verify the delimiter
        List actual = new ArrayList();

        statement = DBPersistenceHelper.executeSQL(connection,
            "SELECT DELIMITER_ENTRY FROM DELIMITER WHERE DOCUMENT_ID = ?", new Object[] {stringId});

        try {
            // Read all delimiters belonging to this document index
            ResultSet rs = statement.getResultSet();

            while (rs.next()) {
                actual.add(rs.getString(1));
            }
        } finally {
            statement.close();
        }

        assertEquals("The delimiters should be persisted.", Arrays.asList(sourceId.getDelimiters()), actual);

        // Verify the words and word positions. They are verified by recontructing the word map
        Map map = new HashMap();

        statement = DBPersistenceHelper.executeSQL(connection,
            "SELECT WORD_ENTRY, WORD_ID FROM WORD WHERE DOCUMENT_ID = ?", new Object[] {stringId});

        try {
            ResultSet rs = statement.getResultSet();

            // Extract every word
            while (rs.next()) {
                // Extract the position of one word
                List positions = new ArrayList();

                PreparedStatement posState = DBPersistenceHelper.executeSQL(connection,
                    "SELECT WORD_POSITION FROM WORD_POSITION WHERE DOCUMENT_ID = ? AND WORD_ID = ?", new Object[] {
                        stringId, rs.getObject(2)});

                try {
                    ResultSet posRs = posState.getResultSet();

                    while (posRs.next()) {
                        positions.add(new Integer(posRs.getInt(1)));
                    }
                } finally {
                    posState.close();
                }

                map.put(collator.getCollationKey(rs.getString(1)), positions);
            }
        } finally {
            statement.close();
        }

        assertEquals("The words and word positions should be persisted.", wordMap, map);
    }

    /**
     * Tests <code>getDocumentIndex(WordSourceId)</code> when a word position in the word data is a negative number.
     * Expect IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentIndexWordSourceIdPositionNegative() throws Exception {
        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US);
        String stringId = IndexPersistenceHelper.generateId(id);

        // Add a document index with no words
        persistence.addDocumentIndex(new DocumentIndex(id, Collections.EMPTY_MAP));

        // Update the document index, so that a position is not a number
        DBPersistenceHelper.executeSQL(connection,
            "INSERT INTO WORD (DOCUMENT_ID, WORD_ID, WORD_ENTRY) VALUES (?, ?, ?)",
            new Object[] {stringId, new Integer(1), "word"}).close();
        DBPersistenceHelper.executeSQL(connection,
            "INSERT INTO WORD_POSITION (DOCUMENT_ID, WORD_ID, WORD_POSITION)" + "VALUES (?, ?, ?)",
            new Object[] {stringId, new Integer(1), new Integer(-1)}).close();

        try {
            persistence.getDocumentIndex(id);
            fail("The word position is a negative number, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getDocumentIndex(WordSourceId)</code> when a word position in the word data is a negative number.
     * Expect IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentIndexWordSourceIdNoPosition() throws Exception {
        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US);
        String stringId = IndexPersistenceHelper.generateId(id);

        // Add a document index with no words
        persistence.addDocumentIndex(new DocumentIndex(id, Collections.EMPTY_MAP));

        // Update the document index, so that a word has no position
        DBPersistenceHelper.executeSQL(connection,
            "INSERT INTO WORD (DOCUMENT_ID, WORD_ID, WORD_ENTRY) VALUES (?, ?, ?)",
            new Object[] {stringId, new Integer(1), "word"}).close();

        try {
            persistence.getDocumentIndex(id);
            fail("The word position is missing, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getDocumentIndex(WordSourceId)</code> when a word appears twice in the WORD table. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentIndexWordSourceIdWordDuplicate() throws Exception {
        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US);
        String stringId = IndexPersistenceHelper.generateId(id);

        // Add a document index with no words
        persistence.addDocumentIndex(new DocumentIndex(id, Collections.EMPTY_MAP));

        // Update the document index, so that a word appears twice
        DBPersistenceHelper.executeSQL(connection,
            "INSERT INTO WORD (DOCUMENT_ID, WORD_ID, WORD_ENTRY) VALUES (?, ?, ?)",
            new Object[] {stringId, new Integer(1), "word"}).close();
        DBPersistenceHelper.executeSQL(connection,
            "INSERT INTO WORD_POSITION (DOCUMENT_ID, WORD_ID, WORD_POSITION)" + "VALUES (?, ?, ?)",
            new Object[] {stringId, new Integer(1), new Integer(0)}).close();
        DBPersistenceHelper.executeSQL(connection,
            "INSERT INTO WORD (DOCUMENT_ID, WORD_ID, WORD_ENTRY) VALUES (?, ?, ?)",
            new Object[] {stringId, new Integer(2), "wORd"}).close();
        DBPersistenceHelper.executeSQL(connection,
            "INSERT INTO WORD_POSITION (DOCUMENT_ID, WORD_ID, WORD_POSITION)" + "VALUES (?, ?, ?)",
            new Object[] {stringId, new Integer(2), new Integer(1)}).close();

        try {
            persistence.getDocumentIndex(id);
            fail("A word appears twice in the WORD table, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>getDocumentIndex(WordSourceId)</code> when there is no word in the document index. A
     * document index with an empty word map should be read.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentIndexWordSourceIdEmptyAccuracy() throws Exception {
        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US);

        // Persist the document index
        persistence.addDocumentIndex(new DocumentIndex(id, Collections.EMPTY_MAP));

        // Load the document index
        DocumentIndex actual = persistence.getDocumentIndex(id);

        // Verify
        assertTrue("The loaded word map should be empty.", actual.getWords().isEmpty());
    }

    /**
     * Tests accuracy of <code>removeDocumentIndex(WordSourceId)</code>. All records related to this document index
     * in table DOCUMENT, DELIMITER, WORD, and WORD_POSITION should be deleted.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testRemoveDocumentIndexWordSourceIdAccuracy() throws Exception {
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US);

        // Create the word map
        Map wordMap = new HashMap();
        Collator collator = Collator.getInstance(Locale.US);

        collator.setStrength(Collator.PRIMARY);

        for (int i = 0; i < 6; ++i) {
            List positions = new ArrayList();

            for (int j = 0; j < 10; ++j) {
                positions.add(new Integer(j));
            }

            wordMap.put(collator.getCollationKey("Word" + Integer.toString(i)), positions);
        }

        // Add another word containing illegal XML character
        wordMap.put(collator.getCollationKey("\1"), Collections.singletonList(new Integer(0)));

        // Create the document index
        DocumentIndex index = new DocumentIndex(id, wordMap);

        // Persist the document index with multiple files
        persistence.addDocumentIndex(index);

        // Persist another document index with no word
        persistence.addDocumentIndex(new DocumentIndex(
            new WordSourceId("id1", "class", new String[] {" "}, Locale.PRC), Collections.EMPTY_MAP));

        // Remove the document index
        persistence.removeDocumentIndex(id);

        // Verify records in tables
        assertNoRecord("The index record in DOCUMENT should be deleted.",
            "SELECT * FROM DOCUMENT WHERE DOCUMENT_ID = ?", new Object[] {IndexPersistenceHelper.generateId(id)});
        assertNoRecord("The index records in DELIMITER should be deleted.",
            "SELECT * FROM DELIMITER WHERE DOCUMENT_ID = ?", new Object[] {IndexPersistenceHelper.generateId(id)});
        assertNoRecord("The index records in WORD should be deleted.", "SELECT * FROM WORD WHERE DOCUMENT_ID = ?",
            new Object[] {IndexPersistenceHelper.generateId(id)});
        assertNoRecord("The index records in WORD_POSITION should be deleted.",
            "SELECT * FROM WORD_POSITION WHERE DOCUMENT_ID = ?", new Object[] {IndexPersistenceHelper.generateId(id)});
    }
}



