/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.xml;

import java.io.File;

import com.topcoder.document.index.persistence.IndexPersistence;
import com.topcoder.document.index.persistence.TestHelper;
import com.topcoder.document.index.persistence.impl.PersistenceConfigurationException;
import com.topcoder.util.config.ConfigManager;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

/**
 * Tests functionality and error cases of <code>XmlIndexPersistence</code> class constructors. Since this class is
 * immutable, the constructors are tested using reflection.
 *
 * @author TCSDEVELOPER
 * @version 1.0
 */
public class XmlIndexPersistenceConstructorTestCase extends TestCase {
    /**
     * Represents the configuration manager used in tests.
     */
    private static final ConfigManager CONFIG = ConfigManager.getInstance();

    /**
     * Represents the namespace used in tests.
     */
    private static final String NAMESPACE = "Test";

    /**
     * Aggragates all tests in this class.
     *
     * @return test suite aggragating all tests.
     */
    public static Test suite() {
        return new TestSuite(XmlIndexPersistenceConstructorTestCase.class);
    }

    /**
     * Sets up the test environment. The test namespace is removed from the configuration.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    protected void setUp() throws Exception {
        if (CONFIG.existsNamespace(NAMESPACE)) {
            CONFIG.removeNamespace(NAMESPACE);
        }
    }

    /**
     * Cleans up the test environment. The test namespace is removed from the configuration.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    protected void tearDown() throws Exception {
        if (CONFIG.existsNamespace(NAMESPACE)) {
            CONFIG.removeNamespace(NAMESPACE);
        }
    }

    /**
     * Tests <code>XmlIndexPersistence(String)</code> when the given namespace is <code>null</code>. Expect
     * IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testXmlIndexPersistenceStringNull() throws Exception {
        try {
            new XmlIndexPersistence(null);
            fail("The namespace is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>XmlIndexPersistence(String)</code> when the given namespace is empty string. Expect
     * IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testXmlIndexPersistenceStringEmpty() throws Exception {
        try {
            new XmlIndexPersistence("  ");
            fail("The namespace is empty string, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>XmlIndexPersistence(String)</code> when the given namespace is missing. Expect
     * IllegalArgumentException.
     */
    public void testXmlIndexPersistenceStringNamespaceMissing() {
        try {
            new XmlIndexPersistence("Missing");
            fail("The namespace is missing in configuration, should throw PersistenceConfigurationException.");
        } catch (PersistenceConfigurationException e) {
            // good
        }
    }

    /**
     * Tests <code>XmlIndexPersistence(String)</code> when the persistence base directory is missing. Expect
     * PersistenceConfigurationException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testXmlIndexPersistenceStringDirMissing() throws Exception {
        CONFIG.add("XmlConfig/DirMissing.xml");

        try {
            new XmlIndexPersistence(NAMESPACE);
            fail("The persistence base directory is missing, should throw PersistenceConfigurationException.");
        } catch (PersistenceConfigurationException e) {
            // good
        }
    }

    /**
     * Tests <code>XmlIndexPersistence(String)</code> when the persistence base directory is not an existing
     * directory. Expect PersistenceConfigurationException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testXmlIndexPersistenceStringDirNotExist() throws Exception {
        CONFIG.add("XmlConfig/DirNotExist.xml");

        try {
            new XmlIndexPersistence(NAMESPACE);
            fail("The persistence base directory does not exist, should throw PersistenceConfigurationException.");
        } catch (PersistenceConfigurationException e) {
            // good
        }
    }

    /**
     * Tests <code>XmlIndexPersistence(String)</code> when the persistence base directory is empty string. Expect
     * PersistenceConfigurationException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testXmlIndexPersistenceStringDirEmpty() throws Exception {
        CONFIG.add("XmlConfig/DirEmpty.xml");

        try {
            new XmlIndexPersistence(NAMESPACE);
            fail("The persistence base directory is empty string, should throw PersistenceConfigurationException.");
        } catch (PersistenceConfigurationException e) {
            // good
        }
    }

    /**
     * Tests <code>XmlIndexPersistence(String)</code> when the file size limit is missing. Expect
     * PersistenceConfigurationException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testXmlIndexPersistenceStringLimitMissing() throws Exception {
        CONFIG.add("XmlConfig/LimitMissing.xml");

        try {
            new XmlIndexPersistence(NAMESPACE);
            fail("The file size limit is missing, should throw PersistenceConfigurationException.");
        } catch (PersistenceConfigurationException e) {
            // good
        }
    }

    /**
     * Tests <code>XmlIndexPersistence(String)</code> when the file size limit is not a number. Expect
     * PersistenceConfigurationException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testXmlIndexPersistenceStringLimitNotNumber() throws Exception {
        CONFIG.add("XmlConfig/LimitNotNumber.xml");

        try {
            new XmlIndexPersistence(NAMESPACE);
            fail("The file size limit is not a number, should throw PersistenceConfigurationException.");
        } catch (PersistenceConfigurationException e) {
            // good
        }
    }

    /**
     * Tests <code>XmlIndexPersistence(String)</code> when the file size limit is 0. Expect
     * PersistenceConfigurationException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testXmlIndexPersistenceStringLimitZero() throws Exception {
        CONFIG.add("XmlConfig/LimitZero.xml");

        try {
            new XmlIndexPersistence(NAMESPACE);
            fail("The file size limit is 0, should throw PersistenceConfigurationException.");
        } catch (PersistenceConfigurationException e) {
            // good
        }
    }

    /**
     * Tests <code>XmlIndexPersistence(String)</code> when the file size safety margin is missing. Expect
     * PersistenceConfigurationException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testXmlIndexPersistenceStringThresholdMissing() throws Exception {
        CONFIG.add("XmlConfig/ThresholdMissing.xml");

        try {
            new XmlIndexPersistence(NAMESPACE);
            fail("The file size safety margin is missing, should throw PersistenceConfigurationException.");
        } catch (PersistenceConfigurationException e) {
            // good
        }
    }

    /**
     * Tests <code>XmlIndexPersistence(String)</code> when the file size safety margin is not a number. Expect
     * PersistenceConfigurationException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testXmlIndexPersistenceStringThresholdNotNumber() throws Exception {
        CONFIG.add("XmlConfig/ThresholdNotNumber.xml");

        try {
            new XmlIndexPersistence(NAMESPACE);
            fail("The file size safety margin is not a number, should throw PersistenceConfigurationException.");
        } catch (PersistenceConfigurationException e) {
            // good
        }
    }

    /**
     * Tests <code>XmlIndexPersistence(String)</code> when the file size safety margin is negative. Expect
     * PersistenceConfigurationException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testXmlIndexPersistenceStringThresholdNegative() throws Exception {
        CONFIG.add("XmlConfig/ThresholdNegative.xml");

        try {
            new XmlIndexPersistence(NAMESPACE);
            fail("The file size safety margin is negative, should throw PersistenceConfigurationException.");
        } catch (PersistenceConfigurationException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>XmlIndexPersistence(String)</code>. The persistence base directory must be loaded from
     * the configuration. The margin file size is calculated by subtract the file size safety margin from the file size
     * limit. The margin file size is in bytes. These are verified via reflection.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testXmlIndexPersistenceStringAccuracy() throws Exception {
        CONFIG.add("XmlConfig/Valid.xml");

        // Create the instance
        IndexPersistence persistence = new XmlIndexPersistence(NAMESPACE);

        // Verify the fields
        assertEquals("The margin file size should be 1024.", new Long(1024), TestHelper.getField(persistence,
            "fileSizeMargin"));
        assertEquals("The base path should be read.", new File("test_files/Xml"), TestHelper.getField(persistence,
            "persistenceDirPath"));
    }

    /**
     * Tests <code>XmlIndexPersistence(int, int, String)</code> when the file size limit is 0. Expect
     * IllegalArgumentException.
     */
    public void testXmlIndexPersistenceIntIntStringLimitZero() {
        try {
            new XmlIndexPersistence(0, 0, "test_files/Xml");
            fail("The file size limit is 0, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>XmlIndexPersistence(int, int, String)</code> when the file size threshold is negative. Expect
     * IllegalArgumentException.
     */
    public void testXmlIndexPersistenceIntIntStringThresholdNegative() {
        try {
            new XmlIndexPersistence(1, -1, "test_files/Xml");
            fail("The file size threshold is negative, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>XmlIndexPersistence(int, int, String)</code> when the persistence base directory is
     * <code>null</code>. Expect IllegalArgumentException.
     */
    public void testXmlIndexPersistenceIntIntStringDirNull() {
        try {
            new XmlIndexPersistence(1, 0, null);
            fail("The persistence base directory is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>XmlIndexPersistence(int, int, String)</code> when the persistence base directory is empty string.
     * Expect IllegalArgumentException.
     */
    public void testXmlIndexPersistenceIntIntStringDirEmpty() {
        try {
            new XmlIndexPersistence(1, 0, "  ");
            fail("The persistence base directory is empty string, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>XmlIndexPersistence(int, int, String)</code> when the persistence base directory does not exist.
     * Expect IllegalArgumentException.
     */
    public void testXmlIndexPersistenceIntIntStringDirNotExist() {
        try {
            new XmlIndexPersistence(1, 0, "NotExist");
            fail("The persistence base directory does not exist, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>XmlIndexPersistence(int, int, String)</code>. The persistence base directory must be
     * the same as given in the argument. The margin file size is calculated by subtract the file size safety margin
     * from the file size limit. The margin file size is in bytes. These are verified via reflection.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testXmlIndexPersistenceIntIntStringAccuracy() throws Exception {
        // Create a new instance
        IndexPersistence persistence = new XmlIndexPersistence(1, 1023, new File("test_files/Xml").getAbsolutePath());

        // Verify the fields
        assertEquals("The margin file size should be 1024.", new Long(1024), TestHelper.getField(persistence,
            "fileSizeMargin"));
        assertEquals("The base path should be read.", new File("test_files/Xml").getAbsolutePath(), ((File) TestHelper
            .getField(persistence, "persistenceDirPath")).getAbsolutePath());
    }
}








